# pylint: disable=missing-module-docstring, wrong-import-position
import unittest
import sys
import os
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.encoding.publicfeatures import PublicFeatures
from bad.encoding.fireworkrank import FireworkRank

class ObservationConverterEncodingTest(unittest.TestCase):
    ''' observation converter '''
    def test_convert_public_features_haseverything(self):
        ''' test convert everything '''
        current_player = 1
        observation = {}
        observation['current_player'] = current_player
        observation['max_action'] = 10
        observation['last_action'] = 5

        observation['player_observations'] = {}
        observation['player_observations'][current_player] = {}

        observation['player_observations'][current_player]['life_tokens'] = 3
        observation['player_observations'][current_player]['information_tokens'] = 8

        card1 = {'color': 'G', 'rank': 1}
        cards = []
        cards.append(card1)
        observation['player_observations'][current_player]['observed_hands'] = []
        observation['player_observations'][current_player]['observed_hands'].append(cards)

        observation['player_observations'][current_player]['fireworks'] = {}
        observation['player_observations'][current_player]['fireworks']['B'] = 5
        observation['player_observations'][current_player]['fireworks']['G'] = 5
        observation['player_observations'][current_player]['fireworks']['R'] = 5
        observation['player_observations'][current_player]['fireworks']['W'] = 5
        observation['player_observations'][current_player]['fireworks']['Y'] = 5

        observation['legal_actions'] = np.array([1,2,3])

        expected_lt_left = np.zeros(shape=4, dtype=int)
        expected_lt_left[3] = 1

        expected_ht_left = np.zeros(shape=9, dtype=int)
        expected_ht_left[8] = 1

        expected_fw_red = np.zeros(shape=6, dtype=int)
        expected_fw_red[5] = 1

        expected_fw_green = np.zeros(shape=6, dtype=int)
        expected_fw_green[5] = 1

        expected_fw_blue = np.zeros(shape=6, dtype=int)
        expected_fw_blue[5] = 1

        expected_fw_white = np.zeros(shape=6, dtype=int)
        expected_fw_white[5] = 1

        expected_fw_yellow = np.zeros(shape=6, dtype=int)
        expected_fw_yellow[5] = 1

        expected_cur_player = np.zeros(shape=2, dtype=int)
        expected_cur_player[1] = 1

        expected_last_action = np.zeros(shape=10, dtype=int)
        expected_last_action[5] = 1

        public_features = PublicFeatures(observation)

        #self.assertTrue(np.array_equal(public_features.life_tokens_left, \
        #    expected_lt_left))
        #self.assertTrue(np.array_equal(public_features.hint_tokens_left, \
        #    expected_ht_left))

        self.assertTrue(np.array_equal(public_features.firework.red, \
            expected_fw_red))
        #self.assertTrue(np.array_equal(public_features.firework.green, \
        #    expected_fw_green))
        #self.assertTrue(np.array_equal(public_features.firework.blue, \
        #    expected_fw_blue))
        #self.assertTrue(np.array_equal(public_features.firework.white, \
        #    expected_fw_white))
        self.assertTrue(np.array_equal(public_features.firework.yellow, \
            expected_fw_yellow))

        self.assertTrue(np.array_equal(public_features.current_player, expected_cur_player))

        self.assertTrue(np.array_equal(public_features.last_action, expected_last_action))

    def test_test_convert_public_features_hasnothing(self):
        ''' convert has nothing '''
        current_player = 0
        observation = {}
        observation['current_player'] = current_player
        observation['max_action'] = 10
        observation['last_action'] = 5

        observation['player_observations'] = {}
        observation['player_observations'][current_player] = {}

        observation['player_observations'][current_player]['life_tokens'] = 0
        observation['player_observations'][current_player]['information_tokens'] = 0

        card1 = {'color': 'G', 'rank': 1}
        cards = []
        cards.append(card1)
        observation['player_observations'][current_player]['observed_hands'] = []
        observation['player_observations'][current_player]['observed_hands'].append(cards)

        observation['player_observations'][current_player]['fireworks'] = {}
        observation['player_observations'][current_player]['fireworks']['B'] = 0
        observation['player_observations'][current_player]['fireworks']['G'] = 0
        observation['player_observations'][current_player]['fireworks']['R'] = 0
        observation['player_observations'][current_player]['fireworks']['W'] = 0
        observation['player_observations'][current_player]['fireworks']['Y'] = 0

        observation['legal_actions'] = np.empty(0, int)

        expected_lt_left = np.zeros(shape=4, dtype=int)
        expected_lt_left[0] = 1

        expected_ht_left = np.zeros(shape=9, dtype=int)
        expected_ht_left[0] = 1

        expectedfireworkrankred = np.zeros(shape=6, dtype=int)
        expectedfireworkrankred[0] = 1

        expectedfireworkrankgreen = np.zeros(shape=6, dtype=int)
        expectedfireworkrankgreen[0] = 1

        expectedfireworkrankblue = np.zeros(shape=6, dtype=int)
        expectedfireworkrankblue[0] = 1

        expectedfireworkrankwhite = np.zeros(shape=6, dtype=int)
        expectedfireworkrankwhite[0] = 1

        expectedfireworkrankyellow = np.zeros(shape=6, dtype=int)
        expectedfireworkrankyellow[0] = 1

        expected_cur_player = np.zeros(shape=2, dtype=int)
        expected_cur_player[0] = 1

        expected_last_action = np.zeros(shape=10, dtype=int)
        expected_last_action[5] = 1

        public_features = PublicFeatures(observation)

        #self.assertTrue(np.array_equal(public_features.life_tokens_left, \
        #    expected_lt_left))
        #self.assertTrue(np.array_equal(public_features.hint_tokens_left, \
        #    expected_ht_left))

        self.assertTrue(np.array_equal(public_features.firework.red, \
            expectedfireworkrankred))
        #self.assertTrue(np.array_equal(public_features.firework.green, \
        #    expectedfireworkrankgreen))
        #self.assertTrue(np.array_equal(public_features.firework.blue, \
        #    expectedfireworkrankblue))
        #self.assertTrue(np.array_equal(public_features.firework.white, \
        #    expectedfireworkrankwhite))
        self.assertTrue(np.array_equal(public_features.firework.yellow, \
            expectedfireworkrankyellow))

        self.assertTrue(np.array_equal(public_features.current_player, expected_cur_player))

        self.assertTrue(np.array_equal(public_features.last_action, expected_last_action))

    def test_differentfireworks(self):
        ''' test different fireworks '''
        current_player = 1
        observation = {}
        observation['current_player'] = current_player

        observation['player_observations'] = {}
        observation['player_observations'][current_player] = {}

        observation['player_observations'][current_player]['life_tokens'] = 3
        observation['player_observations'][current_player]['information_tokens'] = 8

        observation['player_observations'][current_player]['fireworks'] = {}
        observation['player_observations'][current_player]['fireworks']['B'] = 1
        observation['player_observations'][current_player]['fireworks']['G'] = 2
        observation['player_observations'][current_player]['fireworks']['R'] = 3
        observation['player_observations'][current_player]['fireworks']['W'] = 4
        observation['player_observations'][current_player]['fireworks']['Y'] = 5

        observation['current_player'] = current_player

        expectedfireworkrankblue = np.zeros(shape=6, dtype=int)
        expectedfireworkrankblue[1] = 1

        expected_fw_green = np.zeros(shape=6, dtype=int)
        expected_fw_green[2] = 1

        expected_fw_red = np.zeros(shape=6, dtype=int)
        expected_fw_red[3] = 1

        expectedfireworkrankwhite = np.zeros(shape=6, dtype=int)
        expectedfireworkrankwhite[4] = 1

        expectedfireworkrankyellow = np.zeros(shape=6, dtype=int)
        expectedfireworkrankyellow[5] = 1

        firework = FireworkRank(observation)

        self.assertTrue(np.array_equal(firework.red, expected_fw_red))
        #self.assertTrue(np.array_equal(firework.green, expected_fw_green))
        #self.assertTrue(np.array_equal(firework.blue, expectedfireworkrankblue))
        #self.assertTrue(np.array_equal(firework.white, expectedfireworkrankwhite))
        self.assertTrue(np.array_equal(firework.yellow, expectedfireworkrankyellow))

if __name__ == '__main__':
    unittest.main()



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, consider-using-enumerate, line-too-long, line-too-long

import sys
import os
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.collect_episode_data import CollectEpisodeData
from bad.collect_game_result import CollectGameResult
from bad.collect_batch_results import CollectBatchResults
from hanabi_learning_environment import rl_env
from bad.action_network import ActionNetwork
from bad.encoding.observationconverter import ObservationConverter
from bad.set_extra_observation import SetExtraObservation
from bad.reward_to_go_calculation import RewardToGoCalculation
from bad.train_batch_result import TrainBatchResult
from bad.rewards_to_go_calculation_result import RewardsToGoCalculationResult
class TrainBatches:
    """train batch"""
    def __init__(self, network: ActionNetwork, hanabi_environment: rl_env.HanabiEnv, players: int) -> None:
        """init"""
        self.network = network
        self.hanabi_environment = hanabi_environment
        self.players = players

    def collect_data(self, batch_size:int) -> CollectBatchResults:
        """collect data"""
        collect_batch_result = CollectBatchResults()

        while collect_batch_result.get_batch_size() < batch_size:

            hanabi_observation = self.hanabi_environment.reset()
            max_moves: int = self.hanabi_environment.game.max_moves() + 1
            max_actions = max_moves + 1 # 0 index based

            seo = SetExtraObservation()
            seo.set_extra_observation(hanabi_observation, max_moves, max_actions, \
                self.hanabi_environment.state.legal_moves_int())
            observation_converter: ObservationConverter = ObservationConverter()
            self.network.build(observation_converter.convert(hanabi_observation), max_actions)

            ce_data = CollectEpisodeData(hanabi_observation, self.hanabi_environment, self.network)
            game_result: CollectGameResult = ce_data.collect() # hier werden die Daten für eine episode gesammelt

            collect_batch_result.add(game_result)

            #print(f"collected episoden aktionen: {collect_batch_episodes_result.get_batch_size()} von batch size {batch_size}")

        return collect_batch_result

    def calculation(self, collected_data: CollectBatchResults, gamma: float) -> RewardsToGoCalculationResult:
        '''reward to go calculation'''
        reward_to_go_calculation = RewardToGoCalculation(gamma)
        return reward_to_go_calculation.execute(collected_data)

    def backpropagation(self, calc_result: RewardsToGoCalculationResult) -> float:
        '''backpropagation'''
        actions = np.empty(0, int)
        rewards_to_go = np.empty(0, int)
        observation_array_array = []
        baseline =  calc_result.get_baseline()

        for episode_result in calc_result.results:
            for action_index in range(len(episode_result.observation)):
                current_observation = episode_result.observation[action_index].to_one_hot_vec()
                observation_array_array.append(current_observation)
                actions = np.append(actions, episode_result.actions[action_index])
                rewards_to_go = np.append(rewards_to_go, episode_result.rewards_to_go[action_index])

        return self.network.backpropagation(observation_array_array, actions, rewards_to_go, baseline)

    def run(self, batch_size: int, gamma: float) -> TrainBatchResult:
        '''init'''
        print('training')

        print('collecting data')
        collected_data = self.collect_data(batch_size)

        print('reward calculation')
        calculation_result = self.calculation(collected_data, gamma)

        print('backpropagation')
        loss = self.backpropagation(calculation_result)

        return TrainBatchResult(loss, calculation_result.get_game_rewards_sum(), calculation_result.get_rewards_to_go_sum(), calculation_result.get_games_played())



# pylint: disable=missing-module-docstring, wrong-import-position, no-name-in-module, too-few-public-methods, too-many-instance-attributes, too-many-arguments, no-value-for-parameter, wrong-import-order

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from hanabi_learning_environment.rl_env import HanabiEnv
from bad.constants import Constants
from bad.encoding.observation import Observation
from bad.belief.public_belief_player import PublicBeliefPlayer
from bad.belief.hint_matrix_global import HintMatrix
from bad.belief.ftpubvec import RemaingCards
from bad.belief.likelihood_global import Likelihood

class PublicBelief(list):
    """Init"""
    def __init__(self, hanabi_env: HanabiEnv, constants: Constants,
                 action_network, pre_observation_enc: Observation = None,
                 last_action = None, pre_hanabi_env: HanabiEnv = None,
                 public_belief = None):
        """Initialize / Update the PublicBelief
        - create
            RemaningCard (FtpubVec in Paper)
            HintMatrix
            likelihoof

        Creating the publicBelief the first time the parameter
            observation, last_action, pre_hanabi_env and public_belief are None!!!

        Args:
            constants (HanabiEnv): Hanabi environment
            act_network (ActionNetwork): Network that predicts the last_action
            last_act (_type_, None): the last action taken by the agent
            pre_observation (Observation): Observation from the previous step
            pub_belief (Likelihood, None): current public_belief
            pre_hint_matrix (HintMatrix, None): Hint matrix from the previous step
        Returns:
            likelihood (Likelihood): Likelihood of the current step
        """

        # Init rem_cards,hint_matrix and hanabi_env due to dubug purpose
        # Init rem_cards,hint_matrix and hanabi_env due to dubug purpose
        self.rem_cards = RemaingCards(hanabi_env)
        self.hint_matrix = HintMatrix(constants, self.rem_cards)
        self.likelihood = Likelihood(constants, action_network, pre_hanabi_env,
                                     pre_observation_enc, last_action,
                                     public_belief)

        # Init hanabi_env due to dubug purpose
        self.hanabi_env = hanabi_env


        super().__init__(self.init(constants, self.rem_cards,
                         self.hint_matrix, self.likelihood))

    def init(self, constants, rem_cards,
             hint_matrix, likelihood):
        """Init"""

        pub_belf = [PublicBeliefPlayer(constants, idx_ply, rem_cards,
                    hint_matrix[idx_ply], likelihood[idx_ply])
                    for idx_ply in range(constants.num_ply)]
        return pub_belf



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.belief.public_belief_player import PublicBeliefPlayer
from bad.encoding.public_belief_hand_card_enc import PublicBeliefHandCardEnc

class PublicBeliefPlayerEnc():
    '''PublicBeliefPlayerEnc'''
    def __init__(self, public_belf_player: PublicBeliefPlayer):
        '''init'''
        self.hand_card1 = PublicBeliefHandCardEnc(public_belf_player[0])
        self.hand_card2 = PublicBeliefHandCardEnc(public_belf_player[1])
        self.hand_card3 = PublicBeliefHandCardEnc(public_belf_player[2])
        self.hand_card4 = PublicBeliefHandCardEnc(public_belf_player[3])
        self.hand_card5 = PublicBeliefHandCardEnc(public_belf_player[4])



# pylint: disable=missing-module-docstring, wrong-import-position, no-name-in-module, unused-variable, unused-variable, line-too-long, ungrouped-imports
import os
import random
import sys
import numpy as np
import tensorflow as tf

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from hanabi_learning_environment import pyhanabi, rl_env
from bad.action_network import ActionNetwork
from bad.self_play import SelfPlay
from bad.train_epoch import TrainEpoch
from bad.constants import Constants

def main() -> None:
    '''main'''
    seed = 42
    tf.random.set_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    tf.keras.utils.set_random_seed(seed)  # sets seeds for base-python, numpy and tf
    tf.config.experimental.enable_op_determinism()

    batch_size: int = 1000
    epoch_size: int = 1

    episodes_running: int = 10
    gamma: float = 1.0

    model_path = 'model'


    print(f'welcome to bad agent with tf version: {tf.__version__}')
    print(f'running {episodes_running} episodes')

    constants = Constants()
    players = 2
    hanabi_environment = rl_env.make(constants.environment_name, players, pyhanabi.AgentObservationType.SEER)
    network: ActionNetwork = ActionNetwork(model_path)

    #if os.path.exists(model_path):
    #    network.load()

    train_epoch = TrainEpoch(network, hanabi_environment, players)

    for epoch in range(epoch_size):
        print('')
        print(f'running epoch: {epoch}')

        result = train_epoch.train(batch_size, gamma)
        print(f"epoch reward: {result.reward / result.games_played}")

    #network.save()

    self_play = SelfPlay(network)
    self_play.run(episodes_running)

    print("finish with everything")


if __name__ == "__main__":
    main()



# pylint: disable=missing-module-docstring too-few-public-methods, pointless-string-statement,wrong-import-position, fixme, broad-exception-raised
import sys
import os

import numpy as np
import tensorflow_probability as tfp

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.bayesian_action_result import BayesianActionResult


class BayesianAction:
    '''Bayesian Action'''
    def __init__(self, actions: np.ndarray) -> None:
        self.actions = actions
        if len(actions) == 0:
            print()
            raise Exception('no actions')

    def sample_action(self) -> BayesianActionResult:
        '''returns a choice'''
        all_action_probs_distribution = tfp.distributions.Categorical(probs=self.actions)
        sampled_action:int = int(all_action_probs_distribution.sample().numpy())
        return BayesianActionResult(sampled_action)

    def get_action(self, legal_moves:np.ndarray) -> BayesianActionResult:
        '''returns a choice'''
        legal_actions_int = legal_moves.tolist()
        all_action_probs = self.actions.copy()

        if len(legal_actions_int) == len(all_action_probs):
            raise Exception('no legal moves left')

        all_action_probs_distribution = tfp.distributions.Categorical(probs=all_action_probs)
        probs_distribution_as_numpy = all_action_probs_distribution.probs.numpy()
        done = False
        while not done:
            sampled_action:int = int(np.argmax(probs_distribution_as_numpy, axis=0))
            done: bool = legal_actions_int.count(sampled_action) > 0
            if not done:
                probs_distribution_as_numpy[sampled_action] = -float('inf')

        return BayesianActionResult(sampled_action)



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, unused-variable, no-value-for-parameter, no-value-for-parameter

import os
import sys

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.belief.build_hanabi_env import get_hanabi_env
from bad.belief.ftpubvec import RemaingCards


class HintMatrixHandCard(dict):
    '''hint matrix hand card'''
    def __init__(self, constants, rem_cards: RemaingCards, idx_ply: int,
                 idx_card: int):
        '''init'''
        self.idx_ply = idx_ply
        self.idx_card = idx_card
        super().__init__(self.__init(constants, rem_cards))

    def __init_start_condition(self, constants) -> dict:
        """Initializes the HintMatrixHandCard first time"""
        hint_matrix_hand_card = {}

        list_one = [1 for _ in range(constants.num_ranks + 1)]
        for color in constants.colors:
            hint_matrix_hand_card.update({color: list_one.copy()})

        return hint_matrix_hand_card

    def __init(self, constants, rem_cards: RemaingCards) -> dict:
        """Updates the HintMatrixHandCard based on the rem_cards"""
        hint_matrix_hand_card = self.__init_start_condition(constants)

        for color, rem_card_per_color in rem_cards.items():
            for rank in range(constants.max_rank + 1):
                rem_card_per_card = rem_card_per_color[rank]

                # Wenn es keine Karten mehr gibt kann diese auch
                # nicht in der Hand sein
                if rem_card_per_card == 0:
                    hint_matrix_hand_card[color][rank] = 0

        return hint_matrix_hand_card


def main():
    '''main'''
    hanabi_env = get_hanabi_env()
    observation = hanabi_env['player_observations'][0]
    rem_cards = RemaingCards(observation)
    hint_matrix_hand_card = HintMatrixHandCard(observation, rem_cards)
    print()


if __name__ == "__main__":
    main()



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, consider-using-enumerate, line-too-long, line-too-long, too-many-function-args

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.action_network import ActionNetwork
from bad.train_batches import TrainBatches
from bad.train_epoch_result import TrainEpochResult
from hanabi_learning_environment.rl_env import HanabiEnv

class TrainEpoch:
    """train epoch"""
    def __init__(self, network: ActionNetwork, hanabi_environment: HanabiEnv, players:int) -> None:
        """init"""
        self.network = network
        self.train_batches = TrainBatches(self.network, hanabi_environment, players)

    def train(self, batch_size: int, gamma: float) -> TrainEpochResult:
        """train"""
        result = self.train_batches.run(batch_size=batch_size, gamma=gamma)
        return TrainEpochResult(result.loss, result.game_reward, result.games_played)



# pylint: disable=missing-module-docstring too-few-public-methods, pointless-string-statement, fixme


class BayesianActionResult:
    '''BayesianAction result'''
    def __init__(self, sampled_action:int) -> None:
        self.sampled_action = sampled_action



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.encoding.observation import Observation

class ObservationConverter:
    ''' converting hanabi observation to vector observation '''

    def convert(self, observation: dict) -> Observation:
        ''' one hot encoding '''
        return Observation(observation)



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.game_buffer import GameBuffer


class CollectGameResult:
    '''collect episode data result'''
    def __init__(self, buffer: GameBuffer) -> None:
        ''''init'''
        self.buffer = buffer



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, line-too-long

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.encoding.observation import Observation


class RewardsCalculationResult:
    '''RewardsToGoEpisodeCalculationResult'''
    def __init__(self) -> None:
        self.rewards_to_go: list[float] = []
        self.game_rewards: list[float] = []
        self.actions: list[int] = []
        self.observation: list[Observation] = []

    def append(self, action: int, reward_to_go: float, game_rewards:float, observation: Observation) -> None:
        '''add'''
        self.actions.append(action)
        self.rewards_to_go.append(reward_to_go)
        self.game_rewards.append(game_rewards)
        self.observation.append(observation)



# pylint: disable=missing-module-docstring, wrong-import-position, unused-variable, unused-argument, not-callable, invalid-name, fixme, unreachable, line-too-long, consider-using-enumerate, too-many-locals
import sys
import os
import numpy as np
import tensorflow as tf

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.encoding.observation import Observation
from bad.bayesian_action import BayesianAction
from bad.action_provider import ActionProvider
from bad.baseline import Baseline


class ActionNetwork(ActionProvider):
    ''' action network '''

    def __init__(self, path) -> None:
        self.model = None
        self.path = path
        self.optimizer = tf.keras.optimizers.Adam(learning_rate=0.0001)

    def load(self, ) -> None:
        """load"""
        self.model = tf.keras.models.load_model(self.path)

    def save(self):
        """save"""
        self.model.save(self.path)

    def build(self, observation: Observation, max_action: int, \
              public_belief = None) -> None:
        '''build'''
        if self.model is None:
            shape = observation.to_one_hot_vec().shape

            self.model = tf.keras.Sequential([
                tf.keras.Input(shape=shape, name="input"),
                tf.keras.layers.Dense(384, activation="relu", name="layer1"),
                tf.keras.layers.Dense(384, activation="relu", name="layer2"),
                tf.keras.layers.Dense(max_action, activation='softmax', name='Output_Layer')
            ])
            self.model.compile(loss='categorical_crossentropy', optimizer=self.optimizer)

    def print_summary(self):
        '''print summary'''
        self.model.summary()

    def get_model_input(self, observation: Observation, publicBelief=None):
        '''get model input'''
        network_input = observation.to_one_hot_vec()

        # Input muss noch angepasst werden
        # network_input = publicBelief.to_one_hot_vec() + observation.to_one_hot_vec()

        reshaped = tf.reshape(network_input, [1, network_input.shape[0]])
        return reshaped

    def get_action(self, observation: Observation, legal_moves_as_int: list, \
                   public_belief = None) -> BayesianAction:
        '''get action'''
        result = self.model(self.get_model_input(observation, public_belief))
        result_list = result.numpy()[0].tolist()
        result_filtered = [elem_in_res if (elem_idx in legal_moves_as_int) else 0
                           for elem_idx, elem_in_res in enumerate(result_list)]
        return BayesianAction(np.array(result_filtered))

    def backpropagation(self, observation, actions: np.ndarray, rewards_to_go: np.ndarray, baseline: Baseline) -> float:
        """train step"""
        model = self.model
        batch_size = len(observation)
        observation_length = len(observation[0])
        observation_tensor = tf.reshape(observation, [batch_size, observation_length])
        rewards_to_go_tensor = tf.convert_to_tensor(rewards_to_go, dtype=float)

        with tf.GradientTape() as tape:
            logits = model(observation_tensor)
            log_logprobs = tf.nn.log_softmax(logits)

            row_indices= tf.range(len(actions))
            indices = tf.transpose([row_indices, actions])
            logprob = tf.gather_nd(log_logprobs, indices)

            loss = -(tf.reduce_mean(logprob * rewards_to_go_tensor))
            print(f'current loss: {loss.numpy()}')

        grads = tape.gradient(loss, model.trainable_variables)
        self.optimizer.apply_gradients(zip(grads, model.trainable_variables))
        return loss.numpy()



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, consider-using-enumerate, line-too-long, line-too-long
class TrainEpochResult:
    """train epoch result"""
    def __init__(self, loss: float, reward: float, games_played: int) -> None:
        self.loss = loss
        self.reward = reward
        self.games_played = games_played



# pylint: disable=missing-module-docstring, wrong-import-position, no-name-in-module, too-few-public-methods, too-many-instance-attributes

class Constants:
    '''constants'''

    def __init__(self) -> None:
        '''init'''
        self.hanabi_env = None
        self.num_ply = None
        self.num_ranks = None
        self.hand_size = None
        self.num_cards_per_rank = None
        self.num_colors = None
        self.colors = None
        self.environment_name = 'Hanabi-Full'

    def update(self, hanabi_env) -> None:
        '''update'''
        # rank infos needs to update munally
        all_colors = ['R', 'Y', 'G', 'W', 'B']
        self.hanabi_env = hanabi_env
        self.num_ply = hanabi_env.game.num_players()
        self.hand_size = hanabi_env.game.hand_size()
        self.num_ranks = hanabi_env.game.num_ranks()
        self.num_colors = hanabi_env.game.num_colors()
        self.colors = [all_colors[i] for i in range(self.num_colors)]
        self.num_cards_per_rank = [hanabi_env.game.num_cards(all_colors.index('R'), rank)
                                   for rank in range(self.num_ranks + 1)]

    def action_int_to_move_type(self, action: int):
        ''''action int to move type'''
        return self.hanabi_env.game.get_move(action)



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods

import numpy as np
import tensorflow as tf

class NumRemCardsToIntConverter:
    ''' rank converter'''
    def convert(self, rem_cards:int) -> np.ndarray:
        '''convert'''
        return tf.keras.utils.to_categorical(rem_cards, num_classes=6, dtype=int)



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, line-too-long, too-many-arguments, unused-variable, pointless-string-statement
from abc import ABC

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.bayesian_action import BayesianAction
from bad.encoding.observation import Observation


class ActionProvider(ABC):
    """action provider"""
    def get_action(self, observation: Observation, legal_moves_as_int: list, \
                   public_belief = None) -> BayesianAction:
        """get action"""
        raise NotImplementedError("Please Implement this method")



# pylint: disable=missing-module-docstring, wrong-import-position, too-many-locals, too-many-statements
import unittest
import sys
import os
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.encoding.card import Card


class TestCards(unittest.TestCase):
    '''test cards'''
    def test_card_encoding(self):
        '''test card encoding'''

        red_card_one: Card = Card('R', 1)
        red_card_two: Card = Card('R', 2)
        red_card_three: Card = Card('R', 3)
        red_card_four: Card = Card('R', 4)
        red_card_five: Card = Card('R', 5)

        yellow_card_one: Card = Card('Y', 1)
        yellow_card_two: Card = Card('Y', 2)
        yellow_card_three: Card = Card('Y', 3)
        yellow_card_four: Card = Card('Y', 4)
        yellow_card_five: Card = Card('Y', 5)

        green_card_one: Card = Card('G', 1)
        green_card_two: Card = Card('G', 2)
        green_card_three: Card = Card('G', 3)
        green_card_four: Card = Card('G', 4)
        green_card_five: Card = Card('G', 5)

        white_card_one: Card = Card('W', 1)
        white_card_two: Card = Card('W', 2)
        white_card_three: Card = Card('W', 3)
        white_card_four: Card = Card('W', 4)
        white_card_five: Card = Card('W', 5)

        blue_card_one: Card = Card('B', 1)
        blue_card_two: Card = Card('B', 2)
        blue_card_three: Card = Card('B', 3)
        blue_card_four: Card = Card('B', 4)
        blue_card_five: Card = Card('B', 5)

        expected_red_card_one_rank = np.zeros(shape=6, dtype=int)
        expected_red_card_one_rank[1] = 1
        expected_red_card_two_rank = np.zeros(shape=6, dtype=int)
        expected_red_card_two_rank[2] = 1
        expected_red_card_three_rank = np.zeros(shape=6, dtype=int)
        expected_red_card_three_rank[3] = 1
        expected_red_card_four_rank = np.zeros(shape=6, dtype=int)
        expected_red_card_four_rank[4] = 1
        expected_red_card_five_rank = np.zeros(shape=6, dtype=int)
        expected_red_card_five_rank[5] = 1

        expected_yellow_card_one_rank = np.zeros(shape=6, dtype=int)
        expected_yellow_card_one_rank[1] = 1
        expected_yellow_card_two_rank = np.zeros(shape=6, dtype=int)
        expected_yellow_card_two_rank[2] = 1
        expected_yellow_card_three_rank = np.zeros(shape=6, dtype=int)
        expected_yellow_card_three_rank[3] = 1
        expected_yellow_card_four_rank = np.zeros(shape=6, dtype=int)
        expected_yellow_card_four_rank[4] = 1
        expected_yellow_card_five_rank = np.zeros(shape=6, dtype=int)
        expected_yellow_card_five_rank[5] = 1

        expected_green_card_one_rank = np.zeros(shape=6, dtype=int)
        expected_green_card_one_rank[1] = 1
        expected_green_card_two_rank = np.zeros(shape=6, dtype=int)
        expected_green_card_two_rank[2] = 1
        expected_green_card_three_rank = np.zeros(shape=6, dtype=int)
        expected_green_card_three_rank[3] = 1
        expected_green_card_four_rank = np.zeros(shape=6, dtype=int)
        expected_green_card_four_rank[4] = 1
        expected_green_card_five_rank = np.zeros(shape=6, dtype=int)
        expected_green_card_five_rank[5] = 1

        expected_white_card_one_rank = np.zeros(shape=6, dtype=int)
        expected_white_card_one_rank[1] = 1
        expected_white_card_two_rank = np.zeros(shape=6, dtype=int)
        expected_white_card_two_rank[2] = 1
        expected_white_card_three_rank = np.zeros(shape=6, dtype=int)
        expected_white_card_three_rank[3] = 1
        expected_white_card_four_rank = np.zeros(shape=6, dtype=int)
        expected_white_card_four_rank[4] = 1
        expected_white_card_five_rank = np.zeros(shape=6, dtype=int)
        expected_white_card_five_rank[5] = 1

        expected_blue_card_blue_rank = np.zeros(shape=6, dtype=int)
        expected_blue_card_blue_rank[1] = 1
        expected_blue_card_two_rank = np.zeros(shape=6, dtype=int)
        expected_blue_card_two_rank[2] = 1
        expected_blue_card_three_rank = np.zeros(shape=6, dtype=int)
        expected_blue_card_three_rank[3] = 1
        expected_blue_card_four_rank = np.zeros(shape=6, dtype=int)
        expected_blue_card_four_rank[4] = 1
        expected_blue_card_five_rank = np.zeros(shape=6, dtype=int)
        expected_blue_card_five_rank[5] = 1

        self.assertTrue(np.array_equal(red_card_one.rank,   expected_red_card_one_rank))
        self.assertTrue(np.array_equal(red_card_two.rank,   expected_red_card_two_rank))
        self.assertTrue(np.array_equal(red_card_three.rank, expected_red_card_three_rank))
        self.assertTrue(np.array_equal(red_card_four.rank,  expected_red_card_four_rank))
        self.assertTrue(np.array_equal(red_card_five.rank,  expected_red_card_five_rank))

        self.assertTrue(np.array_equal(yellow_card_one.rank,   expected_yellow_card_one_rank))
        self.assertTrue(np.array_equal(yellow_card_two.rank,   expected_yellow_card_two_rank))
        self.assertTrue(np.array_equal(yellow_card_three.rank, expected_yellow_card_three_rank))
        self.assertTrue(np.array_equal(yellow_card_four.rank,  expected_yellow_card_four_rank))
        self.assertTrue(np.array_equal(yellow_card_five.rank,  expected_yellow_card_five_rank))

        self.assertTrue(np.array_equal(green_card_one.rank,   expected_green_card_one_rank))
        self.assertTrue(np.array_equal(green_card_two.rank,   expected_green_card_two_rank))
        self.assertTrue(np.array_equal(green_card_three.rank, expected_green_card_three_rank))
        self.assertTrue(np.array_equal(green_card_four.rank,  expected_green_card_four_rank))
        self.assertTrue(np.array_equal(green_card_five.rank,  expected_green_card_five_rank))

        self.assertTrue(np.array_equal(white_card_one.rank,   expected_white_card_one_rank))
        self.assertTrue(np.array_equal(white_card_two.rank,   expected_white_card_two_rank))
        self.assertTrue(np.array_equal(white_card_three.rank, expected_white_card_three_rank))
        self.assertTrue(np.array_equal(white_card_four.rank,  expected_white_card_four_rank))
        self.assertTrue(np.array_equal(white_card_five.rank,  expected_white_card_five_rank))

        self.assertTrue(np.array_equal(blue_card_one.rank,   expected_blue_card_blue_rank))
        self.assertTrue(np.array_equal(blue_card_two.rank,   expected_blue_card_two_rank))
        self.assertTrue(np.array_equal(blue_card_three.rank, expected_blue_card_three_rank))
        self.assertTrue(np.array_equal(blue_card_four.rank,  expected_blue_card_four_rank))
        self.assertTrue(np.array_equal(blue_card_five.rank,  expected_blue_card_five_rank))



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, line-too-long

class TrainBatchResult:
    '''train batch result'''
    def __init__(self, loss: float, game_reward: float, rewards_to_go:float, games_played: int) -> None:
        """init"""
        self.loss = loss
        self.game_reward = game_reward
        self.rewards_to_go = rewards_to_go
        self.games_played = games_played



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.run_eposiode import RunEpisode
from bad.print_episode_selfplay import PrintEpisodeSelfPlay
from bad.print_total_selfplay import PrintTotalSelfPlay
from bad.action_network import ActionNetwork

class SelfPlay:
    '''self play'''
    def __init__(self, network: ActionNetwork) -> None:
        self.network = network

    def run(self, episodes: int) -> None:
        '''self play'''
        print('self play')

        total_reward = 0
        max_reward = 0
        perfect_games = 0

        for episode in range(episodes):

            run_episode = RunEpisode(self.network)
            episode_result = run_episode.run(episode)

            if episode_result.reward > max_reward:
                max_reward = episode_result.reward
            total_reward += episode_result.reward

            if episode_result.reward == 25:
                perfect_games = perfect_games+1

            print_selfplay = PrintEpisodeSelfPlay(episode_result)
            print_selfplay.print()

        print('')
        print_total_selfplay = PrintTotalSelfPlay(episodes, total_reward, max_reward, perfect_games)
        print_total_selfplay.print()



# pylint: disable=missing-module-docstring, wrong-import-position, unused-variable, unused-argument, not-callable, invalid-name, too-few-public-methods

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)


from bad.encoding.colortointconverter import ColorToIntConverter
from bad.encoding.rem_cards_to_int_converter import NumRemCardsToIntConverter


class CardProbabilitiy:
    '''card'''

    def __init__(self, color: str, num_rem_cards: int) -> None:
        rem_cards_converter: NumRemCardsToIntConverter = NumRemCardsToIntConverter()
        self.num_rem_cards = rem_cards_converter.convert(num_rem_cards)
        color_converter: ColorToIntConverter = ColorToIntConverter()
        self.color = color_converter.convert(color)


def main():
    '''main'''
    card: CardProbabilitiy = CardProbabilitiy('R', 1)
    print(card.num_rem_cards)
    print(card.color)


if __name__ == "__main__":
    main()



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, consider-using-enumerate

class PrintTotalSelfPlay:
    '''print total selfplay'''
    def __init__(self, episodes:int, total_reward:int, max_reward: int, perfect_games: int) -> None:
        '''init'''
        self.total_reward = total_reward
        self.max_reward = max_reward
        self.episodes = episodes
        self.perfect_games = perfect_games

    def print(self) -> None:
        '''print'''
        print(f"Total Reward: {self.total_reward}")
        print(f"Max  Reward: {self.max_reward}")
        print(f"Avg. Reward: {format(self.total_reward/(self.episodes+1),'.3f')}")
        print(f"Perfecet Games: {format(self.perfect_games)}")



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods

import os
import sys
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.encoding.card import Card

class Hands():
    '''hand'''
    def __init__(self, observation: dict) -> None:
        current_player:int = observation['current_player']
        self.own_cards = np.empty(0, int)
        own_hand = observation['player_observations'][current_player]['observed_hands'][0]
        for card in own_hand:

            rank = int(card['rank'])
            if rank == -1:
                rank = 0
            color = card['color']

            if color is None:
                color = 'None'

            own_card = Card(color, rank)
            self.own_cards = np.append(self.own_cards, own_card.color)
            self.own_cards = np.append(self.own_cards, own_card.rank)

        self.own_cards = self.adjust_card_length(self.own_cards)

        self.other_cards = np.empty(0, int)
        other_hands = observation['player_observations'][current_player]['observed_hands'][1:4]
        for other_hand in other_hands:
            for card in other_hand:
                other_card = Card(card['color'], card['rank'])
                self.other_cards = np.append(self.other_cards, other_card.color)
                self.other_cards = np.append(self.other_cards, other_card.rank)

        self.other_cards = self.adjust_card_length(self.other_cards)

    def adjust_card_length(self, cards: np.array) -> np.array:
        '''adjust card length'''
        # in case the agent can not provide all the cards anymore
        # does not seem to be mentioned in the paper
        fill_with_zeros: int = 60 - len(cards)
        if fill_with_zeros > 0:
            cards = np.append(cards, np.zeros((fill_with_zeros, ), dtype=int))
        return cards


if __name__ == "__main__":
    print("Test")



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, no-method-argument, unnecessary-pass, consider-using-enumerate, too-many-function-args, too-many-arguments

import os
import sys

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)


from bad.belief.public_belief_hand_card import PublicBelfHandCard
from bad.belief.ftpubvec import RemaingCards
from bad.belief.hint_matrix_player import HintMatrixPlayer
from bad.belief.likelihood_player import LikelihoodPlayer


class PublicBeliefPlayer(list):
    '''public belief'''
    def __init__(self, constants, idx_ply: int, rem_cards: RemaingCards,
                 hint_matrix_ply: HintMatrixPlayer, likelihood_ply: LikelihoodPlayer):
        '''init'''
        self.idx_ply = idx_ply
        super().__init__(self.init(constants, rem_cards,
                                   hint_matrix_ply, likelihood_ply))

    def init(self, constants, rem_cards: RemaingCards,
             hint_matrix_ply: HintMatrixPlayer,
             likelihood_ply: LikelihoodPlayer) -> list:
        '''init'''

        public_belief_hand_cards = [PublicBelfHandCard(constants, self.idx_ply, idx_card,
                                rem_cards, hint_matrix_ply[idx_card],
                                likelihood_ply[idx_card])
                                for idx_card in range(constants.num_ply)]
        return public_belief_hand_cards
        


# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, no-method-argument, unnecessary-pass, consider-using-enumerate, too-many-function-args
import sys
import os
import time

import numpy as np
currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from hanabi_learning_environment import pyhanabi, rl_env
from bad.action_network import ActionNetwork
from bad.policy import Policy
from bad.set_extra_observation import SetExtraObservation
from bad.run_episode_result import RunEpisodeResult
from bad.bad_agent import BadAgent
from bad.constants import Constants


class RunEpisode:
    '''runs an episode'''
    def __init__(self, network: ActionNetwork) -> None:
        '''init'''
        self.policy = Policy(network)
        players:int = 2
        constants = Constants()
        self.hanabi_environment = rl_env.make(constants.environment_name, players, \
            pyhanabi.AgentObservationType.CARD_KNOWLEDGE)
        constants.update(self.hanabi_environment)
        self.agents = [BadAgent(self.policy, self.hanabi_environment), \
            BadAgent(self.policy, self.hanabi_environment)]
        self.seo = SetExtraObservation()

    def run(self, episode_number: int) -> RunEpisodeResult:
        '''run'''
        print(f"run episode {episode_number}")

        hanabi_observation = self.hanabi_environment.reset()
        # one more move because of no-action move on the beginning
        #fake an action that does not exists
        max_moves: int = self.hanabi_environment.game.max_moves() + 1
        max_actions = max_moves + 1 # 0 index based

        episode_reward = 0
        done = False
        agent_step_times = np.empty(0, float)
        while not done:
            for agent_id in range(len(self.agents)):

                self.seo.set_extra_observation(hanabi_observation, max_moves, max_actions, \
                    self.hanabi_environment.state.legal_moves_int())
                start_time = time.time()
                current_agent = self.agents[agent_id]
                result = current_agent.act(hanabi_observation)
                end_time = time.time()
                agent_step_times = np.append(agent_step_times, end_time - start_time)
                hanabi_observation = result.observation_after_step
                done = result.done
                episode_reward += result.reward
                if done:
                    break

        return RunEpisodeResult(episode_number, episode_reward, self.hanabi_environment, \
            agent_step_times)

# needed ?
#if __name__ == '__main__':
#    runner = RunEpisode()
#    runner.run()



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, too-many-arguments, ungrouped-imports, line-too-long

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.bayesian_action_result import BayesianActionResult
from bad.encoding.observation import Observation
from hanabi_learning_environment.pyhanabi import HanabiMove
from bad.reward_shape import RewardShape

class GameBuffer:
    '''buffer'''
    def __init__(self) -> None:
        self.hanabi_observation: list[dict] = []
        self.observation: list[Observation] = []
        self.bayesian_actions: list[BayesianActionResult] = []
        self.rewards: list[int] = []
        self.moves: list[HanabiMove] = []
        self.reward_shapes: list[RewardShape] =  []

    def append(self, hanabi_observation: dict, observation: Observation, \
    action_result: BayesianActionResult, reward: int, move: HanabiMove, reward_shape: RewardShape) -> None:
        '''add'''
        self.hanabi_observation.append(hanabi_observation)
        self.observation.append(observation)
        self.bayesian_actions.append(action_result)
        self.rewards.append(reward)
        self.moves.append(move)
        self.reward_shapes.append(reward_shape)



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, line-too-long, too-many-arguments, unused-variable, pointless-string-statement, used-before-assignment, no-value-for-parameter

import sys
import os
import copy
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)


from bad.encoding.observation import Observation
from bad.action_provider import ActionProvider

class LikelihoodHandCard(dict):
    '''likely hood hand card'''
    def __init__(self, idx_ply: int, idx_card: int, constants, observation: Observation,
                 act_network: ActionProvider, last_act: int = None, pub_belief = None, pre_hanabi_env = None):
        """Initialize / Update the likelihood based on the action, observation and action network
           By initializing the likelihood, the first time last_act,pub_belief and pre_hanabi_env are None

        Args:
            idx_ply: Index Player curresondes to this HandCard 
            idx_card: Card Index from HandCard
            constants (HanabiEnv): Hanabi environment 
            observation (Observation): Part of Input from Network
            act_network (ActionNetwork): Network that predicts the action
            last_act (int, None): the last action taken by the agent
            public_belief (PublicBelief, None): Public belief from the previous step
            pre_hanabi_env(HanabiEnviornment, None): Hanabi Enviornment from previous step 

        Returns:
            dict: Likelihood to be a spefic for a hand card 
                  based on action of other player 
        """

        self.idx_ply = idx_ply
        self.idx_card = idx_card
        super().__init__(self.init(constants, observation,
                                   act_network, last_act, pub_belief, pre_hanabi_env))

    def init(self, constants, observation: Observation, act_network: ActionProvider,
             last_act, pub_belief, pre_hanabi_env) -> dict:
        '''init'''

        if (pub_belief is None and last_act is None
                and observation is None):
            '''Initialize the likelihood for the first time'''
            return self.first_init_likelihood(constants)

        # Update the likelihood based on the old likelihood, action, input from network and action network
        return self.update_likelihood(constants, observation, act_network,
                                      last_act, pub_belief, pre_hanabi_env)

    def first_init_likelihood(self, constants) -> dict:
        """Initialize the likelihood for the first time"""

        likelihood_hand_card = {}

        list_one = [1 for _ in range(constants.num_ranks + 1)]
        for color in constants.colors:
            likelihood_hand_card.update({color: list_one.copy()})

        return likelihood_hand_card

    def update_likelihood(self, constants, observation: Observation, network: ActionProvider,
                          last_act, pub_belief, pre_hanabi_env) -> dict:
        """Update the likelihood based on the old likelihood, action, input from network and action network"""

        # Get all possible hand_card combinations
        hand_card_combinations = self.hand_card_combinations(constants)

        # Get the possiblity for hand_card_combinations
        hand_card_combinations_possibility = self.hand_card_combinations_pos(
            act_network_output, last_act)

        # Create observations based on the hand_card_combinations, which are used as an input
        # for the action network to get the action based on the hand_card_combinations
        observations_for_hand_card_combinations = self.create_observations(observation, hand_card_combinations)

        # Get action from network based on hand_card_combinations and observation (input from network)
        act_network_output = self.output_actions_network(
            observations_for_hand_card_combinations, pre_hanabi_env, pub_belief, network)

        # Update the hand_card_combinations_possibility
        # based on the action from network
        # (Eliminate the hand_card_combinations that are not possible)
        hand_card_combinations_possibility = self.update_hand_card_combinations_pos(
            hand_card_combinations_possibility, act_network_output, last_act,
            pub_belief.hint_matrix)

        # Normalize the hand_card_combinations_possibility
        hand_card_combinations_possibility = self.normalize_hand_card_possibility(
            hand_card_combinations_possibility)

        # Get the possiblity for each possible hand_card
        hand_card_possibility = self.hand_card_possibility(
            hand_card_combinations_possibility)

        # Update the likelihood based on the hand_card_combinations_possibility
        likelihood = self.calculate_new_likelihood(
            pub_belief, hand_card_possibility)

        return likelihood

    def create_observations(self, observation: Observation, hand_card_combinations) -> list:
        """Create observations based on the hand_card_combinations, which are used as an input
        for the action network to get the action based on the hand_card_combinations"""

        observations_for_hand_card_combinations = []
        for hand_card_combination in hand_card_combinations:
            observations_for_hand_card_combinations.append(
                self.observation_based_on_hand_card_combination(observation, hand_card_combination))

        return observations_for_hand_card_combinations

    def observation_based_on_hand_card_combination(self, observation: Observation,
                                                   hand_card_combination) -> Observation:
        """Create an observation based on the hand_card_combination and old observation"""
        new_observation = copy.copy(observation)
        other_ply_hand = new_observation['player_observations'][0]['observed_hands'][1]
        for idx_card, card in enumerate(hand_card_combination):
            other_ply_hand[idx_card] = card

        return new_observation

    def hand_card_combinations(self, constants) -> list:
        """Returns all possible hand_card combinations"""

        # Get all possible cards
        card_combinations = self.card_combinations(constants)

        # Build all possible hand_card combinations
        hand_card_combinations = []
        for card1 in card_combinations:
            for card2 in card_combinations:
                for card3 in card_combinations:
                    hand_card_combinations.append([card1, card2, card3])

        return hand_card_combinations

    def card_combinations(self, constants) -> list:
        """Return all possible cards"""
        card_combinations = [{'color': color, 'rank': rank} for color in constants.colors
                             for rank in range(constants.max_rank + 1)]

        return card_combinations

    def output_actions_network(self, observations_for_hand_card_combinations,
                               pre_hanabi_env, pub_belief, network: ActionProvider) -> list:
        """Returns the actions from network based on hand_card_combinations and observation (input from network)"""

        output_actions = []
        for observation in observations_for_hand_card_combinations:
            bad = network.get_action(observation)
            bad_result = bad.decode_action(pre_hanabi_env.state.legal_moves_int(), pub_belief)
            next_action = bad_result.sampled_action
            output_actions.append()

        return output_actions

    def build_observations(self, public_feature_one_hot_sig_vec, hand_card_combis_one_hot) -> list:
        """Create Inputs for network"""

        inputs_network = []
        for hand_card_combi_one_hot in hand_card_combis_one_hot:
            inputs_network.append(np.concatenate(public_feature_one_hot_sig_vec, hand_card_combi_one_hot))

        return inputs_network

    def hand_card_combinations_pos(self, hand_card_combinations, public_belief) -> list:
        """Returns the possiblity for hand_card_combinations"""

        hand_card_combinations_possibility = []
        for card_combination in hand_card_combinations:

            # Get the possiblity for each card
            card1_possiblity = self.card_possibility(
                card_combination[0], public_belief, 0)
            card2_possiblity = self.card_possibility(
                card_combination[1], public_belief, 1)
            card3_possiblity = self.card_possibility(
                card_combination[2], public_belief, 2)

            # Get the possiblity for the hand_card_combination and append it to the list
            hand_card_combinations_possibility.append(
                card1_possiblity * card2_possiblity * card3_possiblity)

        return hand_card_combinations_possibility

    def card_possibility(self, card, public_belief, card_idx):
        """Returns the possibility for a card based on the public belief"""
        card_color = card['color']
        card_rank = card['rank']
        card_possibility = public_belief[self.idx_ply][card_idx][card_color][card_rank]

        return card_possibility

    def update_hand_card_combinations_pos(self, hand_card_combis_pos,
                                          network_output, last_act, hint_matrix) -> list:
        """Update the hand_card_combinations_possibility based on the action from network
           (Eliminate the hand_card_combinations that are not possible)

        Args:
            hand_card_combis_pos (list): Possibility for all combinations of own hand based on public belief
                                           A.k.a. the possibility for the hand card combinations 
            network_output (list): Outputs from the network based on the hand card combinations 
            last_act (_type_): last action with the other action played
            hint_matrix (HintMatrix): 

        Return:
            hand_card_combis_pos(list): Updated Version of possibility for our own hand  
        """

        # Eliminate the hand_card_combinations that are not possible
        # based on the action from network
        hand_card_combis_pos = self.update_hand_card_combinations_pos_based_on_net_output(
            hand_card_combis_pos, network_output, last_act)

        # Eliminate the hand_card_combinations that are not possible based on hint_matrix
        hand_card_combis_pos = self.update_hand_card_combinations_pos_based_on_hint_matrix(
            hand_card_combis_pos, hint_matrix)

        return hand_card_combis_pos

    def update_hand_card_combinations_pos_based_on_net_output(self, hand_card_combinations_possibility,
                                                                  act_network_output, last_act):
        """Eliminate the hand_card_combinations that are not possible based on the action from network"""
        for idx, action in enumerate(act_network_output):
            if action != last_act:
                hand_card_combinations_possibility[idx] = 0

        return hand_card_combinations_possibility

    def update_hand_card_combinations_pos_based_on_hint_matrix(self, hand_card_combinations_possibility,
                                                               hint_matrix):
        """Eliminate the hand_card_combinations that are not possible based on hint_matrix"""

        for idx_hand_card_combination, hand_card_combination in hand_card_combinations_possibility:
            for idx_card, card in enumerate(hand_card_combination):
                card_rank = card['rank']
                card_color = card['color']
                if hint_matrix[self.idx_ply][idx_card][card_color][card_rank] == 0:
                    hand_card_combinations_possibility[idx_hand_card_combination] = 0

        return hand_card_combinations_possibility

    def normalize_hand_card_possibility(self, hand_card_combinations_possibility):
        """Normalize the hand_card_combinations_possibility"""

        my_sum = 0
        for possibility in hand_card_combinations_possibility:
            my_sum += possibility

        for i, possibility in enumerate(hand_card_combinations_possibility):
            hand_card_combinations_possibility[i] = possibility / my_sum

        return hand_card_combinations_possibility

    def hand_card_possibility(self, hand_card_combinations_possibility, hand_cards_combination) -> dict:
        """Get the possiblity for each possible hand_card"""
        all_possible_cards = self.card_combinations()

        # Init dict for hand_card_possibility
        hand_card_possibility = {}
        for card in all_possible_cards:
            hand_card_possibility.update({card: 0})

        # Get for each card the possibility based on the hand_card_combinations_possibility
        for hand_card_combination_pos in hand_card_combinations_possibility:

            # The card possibilty is based on the card_idx
            card = hand_cards_combination[self.idx_card]

            hand_card_possibility[card] += hand_card_combination_pos

        return hand_card_possibility

    def calculate_new_likelihood(self, pub_belief: dict, hand_card_combinations_possibility):
        """Update the likelihood based on the hand_card_combinations_possibility and the old likelihood"""
        new_likelihood = {}
        for color, color_likelihood in pub_belief.likelihood.items():
            new_color_likelihood = [hand_card_combinations_possibility[color][rank] * color_likelihood[rank]
                                    for rank in range(len(color_likelihood))]
            new_likelihood.update({color: new_color_likelihood})

        return new_likelihood



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, unused-variable, too-many-arguments

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.belief.build_hanabi_env import get_hanabi_env

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from hanabi_learning_environment.rl_env import HanabiEnv

class PublicBelfHandCard(dict):
    '''public belief hand card'''
    def __init__(self, constants, idx_ply, idx_card, rem_cards,
                 hint_matrix_hand_card, likelihood_hand_card):
        '''init'''
        self.idx_card = idx_card
        self.idx_ply = idx_ply
        super().__init__(self.__init(constants, rem_cards,
                         hint_matrix_hand_card, likelihood_hand_card))

    def __init(self, constants, rem_cards,
               hint_matrix_hand_card, likelihood_hand_card) -> None:
        """Update PublicBelfHandCard based on the rem_cards and hint_matrix_hand_card"""

        public_belief_hand_card = {}

        for color in constants.colors:
            public_belief_color = [rem_cards[color][rank] * hint_matrix_hand_card[color][rank] *
                                   likelihood_hand_card[color][rank]
                                   for rank in  range(constants.num_ranks + 1)]
            public_belief_hand_card.update({color: public_belief_color})

        # Normalize the public_belief_hand_card
        public_belief_hand_card = self.normalize_pub_belief_hand_card(
                                    constants, public_belief_hand_card)

        return public_belief_hand_card

    def normalize_pub_belief_hand_card(self, constants, public_belief_hand_card: dict) -> dict:
        """Normalize the public_belief_hand_card"""
        my_sum = 0
        for color in public_belief_hand_card:
            my_sum += sum(public_belief_hand_card[color])

        for color in public_belief_hand_card:
            public_belief_hand_card[color] = [public_belief_hand_card[color][rank] / my_sum
                                              for rank in range(constants.num_ranks + 1)]

        return public_belief_hand_card

def main():
    '''main'''
    hanabi_env: HanabiEnv = get_hanabi_env()
    print()


if __name__ == "__main__":
    main()



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods
import string
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.encoding.ranktointconverter import RankToIntConverter
from bad.encoding.colortointconverter import ColorToIntConverter

class Card:
    '''card'''
    def __init__(self, color: string, rank: int) -> None:
        rank_converter: RankToIntConverter = RankToIntConverter()
        self.rank = rank_converter.convert(rank)
        color_converter: ColorToIntConverter = ColorToIntConverter()
        self.color = color_converter.convert(color)



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, too-many-arguments, line-too-long

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.reward_shape_result import RewardShapeResult
from bad.reward_shape import RewardShape

class RewardShapeConverter:
    """reward shape converter"""
    def convert(self, reward_shape: RewardShape) -> RewardShapeResult:
        """convert"""

        lost_one_life_token = 0 if reward_shape.lost_one_life_token is True else 1
        lost_all_life_tokens: float = -50 if reward_shape.lost_all_life_tokens is True else 1
        successfully_played_a_card: float = 5 if reward_shape.successfully_played_a_card is True else 0

        discard: float = 0 if reward_shape.discard is True else 0
        discard_playable: float = 0 if reward_shape.discard_playable is None else -10 if reward_shape.discard_playable is True else 5
        discard_unique: float = 0 if reward_shape.discard_unique is None else -10 if reward_shape.discard_unique is True else 5
        discard_useless: float = 0 if reward_shape.discard_useless is None else 10 if reward_shape.discard_useless is True else -5

        hint: float = 0.1 if reward_shape.hint is True else 0
        play: float = 0.5 if reward_shape.play is True else 0

        return RewardShapeResult(lost_one_life_token= lost_one_life_token,
                                 lost_all_life_tokens= lost_all_life_tokens,
                                 successfully_played_a_card= successfully_played_a_card,
                                 discard= discard,
                                 discard_playable= discard_playable,
                                 discard_unique= discard_unique,
                                 discard_useless= discard_useless,
                                 hint= hint,
                                 play= play
                                 )



# pylint: disable=missing-module-docstring, wrong-import-position, no-member, no-name-in-module, too-few-public-methods, line-too-long, ungrouped-imports, too-many-arguments
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from hanabi_learning_environment import rl_env
from hanabi_learning_environment.pyhanabi import HanabiMove
from bad.action_network import ActionNetwork
from bad.encoding.observationconverter import ObservationConverter
from bad.set_extra_observation import SetExtraObservation
from bad.game_buffer import GameBuffer
from bad.collect_game_result import CollectGameResult

from bad.reward_shape import RewardShape
from bad.encoding.observation import Observation
class CollectEpisodeData:
    '''train episode'''

    def __init__(self, hanabi_observation: dict,
                 hanabi_environment: rl_env.HanabiEnv,
                 network: ActionNetwork) -> None:
        self.hanabi_observation = hanabi_observation
        self.hanabi_environment = hanabi_environment
        self.network = network

    def get_reward_shape(self, next_move: HanabiMove) -> RewardShape:
        '''get reward shape'''
        rewardshape = RewardShape()
        rewardshape.execute(next_move, self.hanabi_environment)
        return rewardshape

    def play_a_hanabi_game(self, buffer: GameBuffer, seo: SetExtraObservation, observation: Observation,
                           number_of_actions:int, observation_converter: ObservationConverter):
        """play a hanabi game"""
        #print("play hanabi game")

        done = False
        while not done:
            legal_moves = self.hanabi_environment.state.legal_moves_int()
            bad = self.network.get_action(observation, legal_moves)
            bad_result = bad.sample_action()
            next_action = bad_result.sampled_action
            hanabi_move = self.hanabi_environment.game.get_move(next_action)
            reward_shape = self.get_reward_shape(hanabi_move)
            observation_after_step, reward, done, _ = self.hanabi_environment.step(next_action)
            buffer.append(self.hanabi_observation, observation, bad_result, reward, hanabi_move, reward_shape)
            seo.set_extra_observation(observation_after_step, next_action, number_of_actions, \
                self.hanabi_environment.state.legal_moves_int())
            observation = observation_converter.convert(observation_after_step)
            self.hanabi_observation = observation_after_step

    def collect(self) -> CollectGameResult:
        """train within an environment"""

        copied_state = self.hanabi_environment.state.copy()

        buffer = GameBuffer()

        self.hanabi_environment.state = copied_state.copy()
        # one more move because of no-action move on the beginning
        # fake an action that does not exists
        max_moves: int = self.hanabi_environment.game.max_moves() + 1
        max_actions = max_moves + 1  # 0 index based

        seo = SetExtraObservation()
        seo.set_extra_observation(self.hanabi_observation, max_moves,
                                  max_actions,
                                  self.hanabi_environment.state.legal_moves_int())

        observation_converter: ObservationConverter = ObservationConverter()
        observation = observation_converter.convert(self.hanabi_observation)

        self.play_a_hanabi_game(buffer, seo, observation, max_actions, observation_converter)

        return CollectGameResult(buffer)



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, too-many-arguments, line-too-long, broad-exception-raised, too-many-instance-attributes

import sys
import os
from typing import Optional

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from hanabi_learning_environment.pyhanabi import HanabiCard, HanabiMove, HanabiMoveType, HanabiState
from hanabi_learning_environment import rl_env

class RewardShape:
    '''reward shape'''
    def __init__(self) -> None:
        self.lost_one_life_token: bool = False
        self.lost_all_life_tokens: bool = False
        self.successfully_played_a_card: bool = False

        self.discard: bool = False
        self.discard_playable: Optional[bool] = None
        self.discard_useless: Optional[bool] = None
        self.discard_unique: Optional[bool] = None

        self.play: bool = False
        self.hint: bool = False

    def get_observation_by_state(self, hanabi_state: HanabiState):
        """get observation by state"""
        return hanabi_state.observation(hanabi_state.cur_player())

    def get_card(self, next_move: HanabiMove, hanabi_state: HanabiState) -> HanabiCard:
        """get card knowledge on discarding"""
        my_dict = next_move.to_dict()
        card_index = int(my_dict["card_index"])
        card:HanabiCard = hanabi_state.player_hands()[hanabi_state.cur_player()][card_index]
        return card

    def is_discard_playable(self, next_move: HanabiMove, hanabi_state: HanabiState) -> bool:
        """execute discard playable"""
        observation = self.get_observation_by_state(hanabi_state)
        card = self.get_card(next_move, hanabi_state)
        return bool(observation.card_playable_on_fireworks(card.color(), card.rank()))

    def is_discard_useless(self, next_move: HanabiMove, hanabi_state: HanabiState) -> bool:
        """is discard useless"""
        observation = self.get_observation_by_state(hanabi_state)
        card = self.get_card(next_move, hanabi_state)
        firework = int(observation.fireworks()[card.color()])
        rank = card.one_plus_rank()
        if rank is None:
            raise Exception("rank must have a value")

        return firework >= rank

    def is_discard_unique(self, next_move: HanabiMove, hanabi_state: HanabiState) -> bool:
        """is discard unique"""
        card = self.get_card(next_move, hanabi_state)
        rank = card.one_plus_rank()
        if rank is None:
            raise Exception("rank must have a value")

        return rank == 5

    def execute(self, next_action:HanabiMove , hanabi_environment: rl_env.HanabiEnv) -> None:
        """execute"""
        hanabi_state = hanabi_environment.state.copy()

        current_life_tokens = int(hanabi_state.life_tokens())
        action_type: HanabiMoveType = next_action.type()

        self.discard = action_type == HanabiMoveType.DISCARD
        self.play = action_type == HanabiMoveType.PLAY
        self.hint = action_type == HanabiMoveType.REVEAL_COLOR | HanabiMoveType.REVEAL_RANK

        if action_type == HanabiMoveType.DISCARD:
            self.discard_playable = self.is_discard_playable(next_action, hanabi_state)
            self.discard_useless = self.is_discard_useless(next_action, hanabi_state)
            self.discard_unique = self.is_discard_unique(next_action, hanabi_state)

        hanabi_state.apply_move(next_action)

        next_life_tokens = int(hanabi_state.life_tokens())

        self.lost_one_life_token = current_life_tokens != next_life_tokens
        self.lost_all_life_tokens = current_life_tokens <= 0
        self.successfully_played_a_card = not self.lost_all_life_tokens



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, line-too-long, invalid-name
import sys
import os
import getopt


currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from hanabi_learning_environment.rl_env import HanabiEnv

from hanabi_learning_environment import pyhanabi
from hanabi_learning_environment import rl_env
from hanabi_learning_environment.agents.random_agent import RandomAgent


def get_hanabi_env()->HanabiEnv:
    '''get_hanabi_env'''

    AGENT_CLASSES = {'HTGSAgent' : RandomAgent}

    flags = {'players': 5, 'num_episodes': 100, 'agent_class': 'RandomAgent'}

    options, arguments = getopt.getopt(sys.argv[1:], '',
                                    ['players=',
                                    'num_episodes=',
                                    'agent_class='])
    new_line = '\n'
    if arguments:
        sys.exit(f"usage: rl_env_example.py [options]{new_line} \
                 --players       number of players in the game.{new_line} \
                '--num_episodes  number of game episodes to run.{new_line} \
                '--agent_class   {{}}".format(' or '.join(AGENT_CLASSES.keys())))

    for flag, value in options:
        flag = flag[2:]  # Strip leading --.
        flags[flag] = type(flags[flag])(value)

    environment = rl_env.make('Hanabi-Full', num_players=flags['players'], agentObservationType = pyhanabi.AgentObservationType.CARD_KNOWLEDGE)


    observations = environment.reset()

    return observations

def main():
    '''main'''
    observation: HanabiEnv = get_hanabi_env()
    print(observation)

if __name__ == "__main__":
    main()



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, too-many-arguments, line-too-long, too-many-instance-attributes
class RewardShapeResult:
    """reward shape result"""
    def __init__(self, lost_one_life_token: float, lost_all_life_tokens: float, successfully_played_a_card: float,
                discard: float, discard_playable: float, discard_unique: float, discard_useless: float,
                hint: float, play: float
                 ) -> None:
        """init"""
        self.lost_one_life_token = lost_one_life_token
        self.lost_all_life_tokens = lost_all_life_tokens
        self.successfully_played_a_card = successfully_played_a_card
        self.discard = discard
        self.discard_playable = discard_playable
        self.discard_unique = discard_unique
        self.discard_unseless = discard_useless
        self.hint = hint
        self.play = play

    def get_sum(self) -> float:
        """sum"""
        return self.lost_all_life_tokens + self.lost_all_life_tokens + self.discard_playable



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods

import sys
import os
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.encoding.public_belief_player_enc import PublicBeliefPlayerEnc

class PublicBeliefGlobalEnc():
    """Inherit from list"""
    def __init__(self, public_belf):
        self.public_belf_player1 = PublicBeliefPlayerEnc(public_belf[0])
        self.public_belf_player2 = PublicBeliefPlayerEnc(public_belf[1])

    def to_one_hot(self):
        """Convert Public Belief to one-hot encoding"""
        result = np.concatenate((
            self.public_belf_player1.hand_card1,
            self.public_belf_player1.hand_card2,
            self.public_belf_player1.hand_card3,
            self.public_belf_player1.hand_card4,
            self.public_belf_player1.hand_card5,
            self.public_belf_player2.hand_card1,
            self.public_belf_player2.hand_card2,
            self.public_belf_player2.hand_card3,
            self.public_belf_player2.hand_card4,
            self.public_belf_player2.hand_card5
             ))
        return result



# pylint: disable=missing-module-docstring, wrong-import-position, wrong-import-order, ungrouped-imports, too-few-public-methods, line-too-long, too-many-arguments

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.action_network import ActionNetwork
from bad.encoding.observation import Observation
from bad.belief.likelihood_player import LikelihoodPlayer

class Likelihood(list):
    '''likely hood'''
    def __init__(self, constants, action_network: ActionNetwork, pre_hanabi_env,
                pre_observation: Observation, last_act, pub_belief)-> None:
        """Initialize / Update the likelihood based on the last_action, observation and last_action network
        By initializing the likelihood, the first time last_act,old_likelihood and last_act are None
        

        Args:
            constants (HanabiEnv): Hanabi environment 
            act_network (ActionNetwork): Network that predicts the last_action
            last_act (int, None): the last action taken by the agent
            pre_observation (Observation): Observation from the previous step
            pub_belief (Likelihood, None): current public_belief
            pre_hint_matrix (HintMatrix, None): Hint matrix from the previous step
        Returns:
            likelihood (Likelihood): Likelihood of the current step
        """
        super().__init__(self.__init(constants, pre_observation,
                                     action_network, last_act, pub_belief, pre_hanabi_env))


    def __init(self, constants, observation: Observation,
                     action_network: ActionNetwork, last_act,
                     pub_belief, pre_hanabi_env) -> list:

        players_hands = [LikelihoodPlayer(constants, idx_ply, observation,
                         action_network, last_act, pub_belief, pre_hanabi_env)
                         for idx_ply in range(constants.num_ply)]

        return players_hands



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, consider-using-enumerate, line-too-long, line-too-long
class Baseline:
    """baseline"""
    def __init__(self, mean:float, std: float) -> None:
        """init"""
        self.mean = mean
        self.std = std



# pylint: disable=missing-module-docstring, wrong-import-position, wrong-import-order, ungrouped-imports, too-few-public-methods, line-too-long, too-many-arguments, no-value-for-parameter

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.belief.ftpubvec import RemaingCards
from bad.belief.hint_matrix_hand_card import HintMatrixHandCard


class HintMatrixPlayer(list):
    '''hint matrix player'''
    def __init__(self, constants, idx_ply: int, rem_cards: RemaingCards):
        self.idx_ply = idx_ply
        super().__init__(self.__init(constants, rem_cards))

    def __init(self, constants, rem_cards: RemaingCards) -> list:
        hint_matrix_player = [HintMatrixHandCard(constants, idx_card, rem_cards)
                              for idx_card in range(constants.num_ply)]
        return hint_matrix_player



# pylint: disable=missing-module-docstring, wrong-import-position, unused-variable, unused-argument, not-callable, invalid-name, too-few-public-methods

import sys
import os
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.belief.public_belief_hand_card import PublicBelfHandCard
from bad.encoding.card_probabilitiy import CardProbabilitiy

class PublicBeliefHandCardEnc():
    '''PublicBeliefHandCardEnc'''
    def __init__(self, public_belf_hand_card: PublicBelfHandCard):
        '''init'''
        colors = ['B', 'G', 'R', 'W', 'Y']
        max_rank = 5

        self.pro_hand_card = np.empty(0,int)

        for color in colors:
            for rank in range(max_rank):
                num_rem_cards: int = public_belf_hand_card[color][rank]
                self.pro_hand_card = np.append(self.pro_hand_card,
                    CardProbabilitiy([color],[num_rem_cards]))



# pylint: disable=missing-module-docstring, wrong-import-position, no-name-in-module, too-few-public-methods, too-many-instance-attributes, too-many-arguments

import sys
import os


currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.action_provider import ActionProvider
from bad.encoding.observation import Observation
from bad.belief.likelihood_hand_card import LikelihoodHandCard


class LikelihoodPlayer(list):
    '''likelihoodplayer'''
    def __init__(self, constants, idx_ply: int, observation: Observation,
                  action_network: ActionProvider, last_act, pub_belief, pre_hanabi_env):
        '''init'''
        self.idx_ply = idx_ply
        super().__init__(self.__init(constants, idx_ply, observation,
                                     action_network, last_act, pub_belief, pre_hanabi_env))

    def __init(self, constants, idx_ply, observation,
               action_network, last_act, pre_pub_belief, pre_hanabi_env) -> list:
        '''init'''
        likelihood_player = [LikelihoodHandCard(idx_ply, idx_card, constants, observation,
                              action_network, last_act, pre_pub_belief, pre_hanabi_env)
                              for idx_card in range(constants.num_cards) ]
        return likelihood_player



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, too-many-arguments
import numpy as np

class SetExtraObservation:
    '''set extra observation'''
    def set_extra_observation(self, hanabi_observation:dict, \
        last_action: int, max_action:int, legal_moves: list) -> None:
        '''set extra observation'''
        hanabi_observation['last_action'] = last_action
        hanabi_observation['max_action'] = max_action

        legal_moves_int = np.empty(0, int)
        for legal_move in legal_moves:
            legal_moves_int = np.append(legal_moves_int, legal_move)

        hanabi_observation['legal_actions'] = legal_moves_int



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, invalid-name
import sys
import os
import numpy as np


currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.rewards_to_go_episode_calculation_result import RewardsCalculationResult
from bad.baseline import Baseline

class RewardsToGoCalculationResult:
    """RewardToGoCalculationResult"""
    def __init__(self, games_played) -> None:
        self.results: list[RewardsCalculationResult] = []
        self.games_played = games_played

    def append(self, result: RewardsCalculationResult):
        '''append'''
        self.results.append(result)

    def get_games_played(self) -> int:
        """get games played"""
        return self.games_played

    def get_batch_size(self) -> int:
        """get n"""
        batch_size: int = 0

        for ep in self.results:
            batch_size += len(ep.actions)

        return batch_size

    def get_rewards_to_go(self) -> np.ndarray:
        """get rewards"""
        total = np.empty(0, float)

        for res in self.results:
            total = np.append(total, res.rewards_to_go)

        return total

    def get_baseline(self) -> Baseline:
        """get baseline"""
        rewards = self.get_rewards_to_go()
        return Baseline(rewards.mean(), rewards.std())

    def get_rewards_to_go_sum(self) -> float:
        """get rewards to go sum"""
        rewards = self.get_rewards_to_go()

        return rewards.sum()

    def get_game_rewards(self) -> np.ndarray:
        """get game rewards"""
        total = np.empty(0, float)

        for res in self.results:
            total = np.append(total, res.game_rewards)

        return total

    def get_game_rewards_sum(self) -> float:
        """get game reward"""
        rewards = self.get_game_rewards()
        return rewards.sum()



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, line-too-long
import os
import sys

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from hanabi_learning_environment.rl_env import HanabiEnv

class RemaingCards(dict):
    """Es handels sich um FtPubVec aus der Arbeit von Foster
       Beinhaltet die Information der remaing Cards in Form eines Dict """
    def __init__(self, hanabi_env: HanabiEnv, constants) -> None:
        super().__init__(self.__init(hanabi_env, constants))

    def init_start_conditions(self, constants)-> dict:
        '''init start conditions'''

        rem_cards = {'B': constants.num_cards_per_ranks.copy(),
                     'G': constants.num_cards_per_ranks.copy(),
                     'R': constants.num_cards_per_ranks.copy(),
                     'W': constants.num_cards_per_ranks.copy(),
                     'Y': constants.num_cards_per_ranks.copy(),}

        return rem_cards

    def __init(self, hanabi_env: HanabiEnv, constants)-> dict:
        """Based on the Public Information we calculate rem_cards
        For each card we find in public information (card_knowledge,
        discard_pile and firework) we reduce max number by one"""
        rem_cards = self.init_start_conditions(constants)

        self.update_based_on_card_knowledge(hanabi_env, rem_cards)
        self.update_based_on_discard_pile(hanabi_env, rem_cards)
        self.update_based_on_firework(hanabi_env, rem_cards)

        return rem_cards

    def update_based_on_firework(self, hanabi_env: HanabiEnv, rem_cards)-> dict:
        """ Update rem_cards based on Firework"""

        # Jede Karte die im Firework liegt kann nicht
        # mehr auf der Hand eines Spieles sein
        firework = hanabi_env['player_observations'][0]['fireworks']
        for color, max_rank in firework.items():
            for rank in range(max_rank):
                rem_cards[color][rank] -= 1

        return rem_cards

    def update_based_on_card_knowledge(self, hanabi_env: HanabiEnv, rem_cards)-> dict:
        """Update rem_cards based on card_knowledge """

        card_knowledge = hanabi_env['player_observations'][0]['card_knowledge']

        # Prüfe alle Karten in card_knowledge
        for player_card_knowledge in card_knowledge:
            for card in player_card_knowledge:

                # Wenn eine Karte vollständig bekannt dann reduziere mc
                # Diese Karte kann ja nicht mehr einer anderen Hand sein
                if (card['rank'] is not None and
                    card['color'] is not None):

                    rem_cards[card['color']][card['rank']] -= 1

        return rem_cards

    def update_based_on_discard_pile(self, hanabi_env: HanabiEnv, rem_cards)-> dict:
        """Update mc based on discard_pile """

        # Jede Karte die im Discard Pile ist kann nicht mehr in der Hand
        # eines anderen Spieler sein
        discard_pile = hanabi_env['player_observations'][0]['discard_pile']
        for card in discard_pile:
            rem_cards[card['color']][card['rank']] -= 1

        return rem_cards



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods
import string
import sys
import os
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)


from bad.encoding.ranktointconverter import RankToIntConverter

class FireworkRank():
    ''' firework rank '''
    def __init__(self, observation: dict) -> None:
        '''init'''
        self.observation = observation
        self.current_player = observation['current_player']

        self.red = self.convert_firework_rank_color('R')
        self.yellow = self.convert_firework_rank_color('Y')
        self.green = self.convert_firework_rank_color('G')
        self.white  = self.convert_firework_rank_color('W')
        self.blue = self.convert_firework_rank_color('B')

    def convert_firework_rank_color(self, color: string) -> np.ndarray:
        ''' convert firework rank color '''
        rank = self.observation['player_observations'][self.current_player]['fireworks'][color]
        rank_converter = RankToIntConverter()
        return rank_converter.convert(rank)



# pylint: disable=missing-module-docstring, wrong-import-position, too-few-public-methods, invalid-name

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.collect_game_result import CollectGameResult


class CollectBatchResults:
    """collect bactch data results"""
    def __init__(self) -> None:
        self.results: list[CollectGameResult] = []

    def add(self, result: CollectGameResult) -> None:
        '''add'''
        self.results.append(result)

    def get_games_played(self) -> int:
        """games played"""
        return len(self.results)

    def get_batch_size(self) -> int:
        """get batch size"""
        batch_size: int = 0

        for ep in self.results:
            batch_size += len(ep.buffer.bayesian_actions)

        return batch_size



# pylint: disable=missing-module-docstring, wrong-import-position, wrong-import-order, ungrouped-imports, too-few-public-methods, line-too-long, too-many-arguments, unused-variable, no-value-for-parameter

import os
import sys


currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
parentPath2 = os.path.dirname(parentPath)
sys.path.append(parentPath2)

from bad.belief.hint_matrix_player import HintMatrixPlayer
from bad.belief.ftpubvec import RemaingCards
from bad.belief.build_hanabi_env import get_hanabi_env



class HintMatrix(list):
    '''hint matrix'''
    def __init__(self, constants, rem_cards: RemaingCards):
        '''init'''
        super().__init__(self.__init(constants, rem_cards))


    def __init(self, constants, rem_cards: RemaingCards) -> list:
        '''init'''
        players_hands = [HintMatrixPlayer(constants, idx_ply,
                         rem_cards) for idx_ply in range(constants.num_ply)]

        return players_hands

def main():
    '''main'''
    hanabi_env = get_hanabi_env()
    observation = hanabi_env['player_observations']
    rem_cards = RemaingCards(hanabi_env)
    hint_matrix = HintMatrix(observation, rem_cards)
    print()

if __name__ == "__main__":
    main()



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, line-too-long, consider-using-enumerate, unused-variable

import sys
import os
import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.collect_batch_results import CollectBatchResults
from bad.rewards_to_go_episode_calculation_result import RewardsCalculationResult
from bad.rewards_to_go_calculation_result import RewardsToGoCalculationResult
from bad.reward_shape_converter import RewardShapeConverter
from bad.game_buffer import GameBuffer

class RewardToGoCalculation:
    ''''calculate reward to go'''
    def __init__(self, gamma: float) -> None:
        self.gamma = gamma

    def calculate(self, buffer: GameBuffer, result: RewardsCalculationResult) -> None:
        ''''calculate episode'''

        reward_shape_converter = RewardShapeConverter()

        for index in range(len(buffer.bayesian_actions)): # über jede aktion (pro spiel)
            reward_shape = reward_shape_converter.convert(buffer.reward_shapes[index])
            # hier rewards verändern
            reward_vom_hanabi_framework = float(np.sum(buffer.rewards[index:]))
            reward_vom_reward_shaping = 0.0 # reward_shape.get_sum()

            reward_to_go = reward_vom_hanabi_framework + reward_vom_reward_shaping
            discounted_reward_to_go = reward_to_go * np.power(self.gamma, index + 1)
            observation = buffer.observation[index]
            bayesian_actions = buffer.bayesian_actions[index]

            result.append(bayesian_actions.sampled_action, discounted_reward_to_go, reward_vom_hanabi_framework, observation)

    def execute(self,collected_batch_results: CollectBatchResults) -> RewardsToGoCalculationResult:
        """execute"""
        episodes_result = RewardsToGoCalculationResult(collected_batch_results.get_games_played())

        for batch_result in collected_batch_results.results:
            reward_calculation_result = RewardsCalculationResult()
            episodes_result.append(reward_calculation_result)

            buffer = batch_result.buffer
            self.calculate(buffer, reward_calculation_result)

        return episodes_result
